from __future__ import annotations

from dataclasses import dataclass, fields
from enum import Enum
from pathlib import Path
from typing import Any, Dict

import torch
from transformers import TrainingArguments

BASE_MODEL = "facebook/wav2vec2-base-960h"
SAMPLING_RATE = 16_000


class TrainingStatus(Enum):
    WAITING = "waiting"
    TRAINING = "training"
    FINISHED = "finished"
    ERROR = "error"


@dataclass
class TrainingOptions:
    """A class representing some commonly changed training options"""

    batch_size: int = 4
    epochs: int = 2
    learning_rate: float = 1e-4
    min_duration: int = 0
    max_duration: int = 60
    word_delimiter_token: str = " "
    test_size: float = 0.2
    freeze_feature_extractor: bool = False

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "TrainingOptions":
        field_names = [field.name for field in fields(TrainingOptions)]
        kwargs = {key: data[key] for key in data if key in field_names}
        return TrainingOptions(**kwargs)

    def to_dict(self) -> Dict[str, Any]:
        return dict(self.__dict__)


@dataclass
class TrainingJob:
    """A class representing a training job for a model"""

    model_name: str
    dataset_name: str
    options: TrainingOptions
    status: TrainingStatus = TrainingStatus.WAITING
    base_model: str = BASE_MODEL
    sampling_rate: int = SAMPLING_RATE

    def to_training_args(self, output_dir: Path, **kwargs) -> TrainingArguments:
        return TrainingArguments(
            output_dir=str(output_dir),
            group_by_length=True,
            num_train_epochs=self.options.epochs,
            per_device_train_batch_size=self.options.batch_size,
            per_device_eval_batch_size=self.options.batch_size,
            gradient_accumulation_steps=2,
            evaluation_strategy="steps",
            fp16=True if torch.cuda.is_available() else False,
            gradient_checkpointing=True,
            learning_rate=self.options.learning_rate,
            weight_decay=0.005,
            save_steps=500,
            eval_steps=500,
            logging_steps=500,
            warmup_steps=1000,
            save_total_limit=2,
            overwrite_output_dir=True,
            do_train=True,
            do_eval=True,
            **kwargs,
        )

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> TrainingJob:
        return TrainingJob(
            model_name=data["model_name"],
            dataset_name=data["dataset_name"],
            options=TrainingOptions.from_dict(data["options"]),
            status=TrainingStatus(data.get("status", TrainingStatus.WAITING)),
            base_model=data.get("base_model", BASE_MODEL),
            sampling_rate=data.get("sampling_rate", SAMPLING_RATE),
        )

    def to_dict(self) -> Dict[str, Any]:
        result = dict(self.__dict__)
        result |= dict(options=self.options.to_dict(), status=self.status.value)
        return result



import re
from typing import List, Optional


def clean_text(
    text: str,
    words_to_remove: Optional[List[str]] = None,
    punctuation_to_explode: str = "",
    punctuation_to_remove: str = "",
) -> str:
    """Cleans the text based on the supplied options.

    Parameters:
        text: The text to clean.
        options: The cleaning options.

    Returns:
        The cleaned text
    """
    words = text.lower().split()

    if words_to_remove is not None:
        words = filter(lambda word: word not in words_to_remove, words)

    if punctuation_to_explode != "":
        words = map(lambda word: explode(word, punctuation_to_explode), words)

    if punctuation_to_remove != "":
        words = map(lambda word: collapse(word, punctuation_to_remove), words)

    result = " ".join(words).strip()
    return remove_consecutive_spaces(result)


def explode(text: str, pattern: str) -> str:
    """Replace occurences of the pattern with spaces within the given text.

    Parameters:
        text: The text to modify.
        pattern: The pattern of characters to replace with spaces.

    Returns:
        The text with instances of the pattern exploded.
    """
    pattern = re.escape(pattern)
    return re.sub(rf"[{pattern}]", " ", text)


def collapse(text: str, pattern: str) -> str:
    """Remove occurences of the pattern within the given text.

    Parameters:
        text: The text to modify.
        pattern: The pattern of characters to remove.

    Returns:
        The text with instances of the pattern removed.
    """
    pattern = re.escape(pattern)
    return re.sub(rf"[{pattern}]", "", text)


def remove_consecutive_spaces(text: str) -> str:
    """Replace consecutive spaces with a single one in some given text.

    Parameters:
        text: The text to modify.

    Returns
        The supplied text with conseucutive spaces reduced to one.
    """
    return re.sub("[ ]+", " ", text)



import os
from pathlib import Path
from typing import Any, Dict, List, Optional

from datasets import Audio, DatasetDict, load_dataset
from transformers import Wav2Vec2Processor

PROCESSOR_COUNT = 4
AUDIO_COLUMN = "audio"
SAMPLING_RATE = 16_000


def create_dataset(
    dataset_path: Path, cache_dir: Optional[Path] = None, test_size: float = 0.2
) -> DatasetDict:
    """Creates a dataset with test/train splits from the data within a given
    directory.

    Parameters:
        dataset_path: The path to the unprocessed dataset files.
        cache_dir: The path to save the processed dataset.
        test_size: The percentage of the dataset to allocate as the test set.

    Returns:
        A dataset dictionary with test and train splits.
    """

    transcript_files = [
        str(dataset_path / file)
        for file in os.listdir(dataset_path)
        if (dataset_path / file).suffix == ".json"
    ]
    # Annoying hack
    if cache_dir is not None:
        cache_dir = str(cache_dir)  # type: ignore

    dataset = load_dataset("json", data_files=transcript_files, cache_dir=cache_dir)  # type: ignore

    # Convert the audio file name column into the matching audio data
    dataset = dataset.rename_column("audio_file", AUDIO_COLUMN)

    def resolve_audio_path(row: Dict[str, Any]) -> Dict[str, Any]:
        row[AUDIO_COLUMN] = str(dataset_path / row[AUDIO_COLUMN])
        return row

    dataset = dataset.map(resolve_audio_path)
    dataset = dataset.cast_column(AUDIO_COLUMN, Audio(sampling_rate=SAMPLING_RATE))

    return dataset["train"].train_test_split(test_size=test_size)  # type: ignore


def prepare_dataset(dataset: DatasetDict, processor: Wav2Vec2Processor) -> DatasetDict:
    """Runs some preprocessing over the given dataset.

    TODO: I'm going to be honest, I have no idea what this does, and need some
    smart ML knight in shining armour to write a propert description.

    Parameters:
        dataset: The dataset to apply the preprocessing
        processor: The processor to apply over the dataset
    """

    def prepare_dataset(batch: Dict) -> Dict[str, List]:
        audio = batch["audio"]

        batch["input_values"] = processor(
            audio["array"], sampling_rate=audio["sampling_rate"]
        ).input_values[0]
        batch["input_length"] = len(batch["input_values"])

        with processor.as_target_processor():
            batch["labels"] = processor(batch["transcript"]).input_ids

        return batch

    return dataset.map(
        prepare_dataset,
        remove_columns=dataset.column_names["train"],
        num_proc=PROCESSOR_COUNT,
    )



from dataclasses import dataclass
from enum import Enum
from typing import Dict


class ElanTierSelector(Enum):
    """A class representing a method of selecting elan tiers"""

    ORDER = "tier_order"
    TYPE = "tier_type"
    NAME = "tier_name"


@dataclass
class ElanOptions:
    """A class representing options for how to extract utterance information
    from an elan file."""

    selection_mechanism: ElanTierSelector
    selection_value: str

    @classmethod
    def from_dict(cls, data: Dict[str, str]) -> "ElanOptions":
        return cls(
            selection_mechanism=ElanTierSelector(data["selection_mechanism"]),
            selection_value=data["selection_value"],
        )

    def to_dict(self) -> Dict[str, str]:
        result = dict(self.__dict__)
        result["selection_mechanism"] = self.selection_mechanism.value
        return result



from elpis.transcriber.results import build_elan, build_text
from elpis.transcriber.transcribe import build_pipeline, transcribe

__all__ = ["build_elan", "build_text", "build_pipeline", "transcribe"]



from pathlib import Path
from typing import Any, Dict, List, Optional

from transformers import AutomaticSpeechRecognitionPipeline as ASRPipeline
from transformers import AutoModelForCTC, AutoProcessor, pipeline

from elpis.models import Annotation

TASK = "automatic-speech-recognition"


def build_pipeline(
    pretrained_location: str,
    processor_location: Optional[str] = None,
    cache_dir: Optional[Path] = None,
) -> ASRPipeline:
    """Builds the pipeline from the supplied pretrained location.

    Parameters:
        pretrained_location: A huggingface model name, or local path to the
            pretrained model.
        cache_dir: The directory in which to store temporary files.

    Returns:
        A pipeline to be used for asr.
    """
    if processor_location is None:
        processor_location = pretrained_location

    processor = AutoProcessor.from_pretrained(processor_location, cache_dir=cache_dir)
    model = AutoModelForCTC.from_pretrained(
        pretrained_location,
        cache_dir=cache_dir,
        pad_token_id=processor.tokenizer.pad_token_id,
    )

    return pipeline(
        task=TASK,
        model=model,
        tokenizer=processor.tokenizer,
        feature_extractor=processor.feature_extractor,
    )  # type: ignore


def transcribe(audio: Path, asr: ASRPipeline) -> List[Annotation]:
    """Transcribes the given audio and gives back the resulting annotations.

    Parameters:
        audio: The path to the audio file to transcribe.
        asr: The automatic speech recognition pipeline.

    Returns:
        A list of the inferred annotations in the given audio.
    """
    preds: Dict[str, Any] = asr(str(audio), return_timestamps="word")  # type: ignore
    chunks = preds["chunks"]

    return list(map(lambda chunk: annotation_from_chunk(chunk, audio), chunks))


def annotation_from_chunk(chunk: Dict[str, Any], audio_file: Path) -> Annotation:
    """Builds an annotation from a chunk.

    Chunks are in the form:
        {"text": "some_text", "timestamp": (start, stop)}

    Parameters:
        chunk: The chunk to convert
        audio_file: The file which the chunk is for.

    Returns:
        The corresponding annotation.
    """
    text = chunk["text"]
    start, stop = chunk["timestamp"]

    return Annotation(
        transcript=text,
        start_ms=int(start * 1000),
        stop_ms=int(stop * 1000) + 1,
        audio_file=audio_file,
    )



from pathlib import Path

from pedalboard.io import ReadableAudioFile, WriteableAudioFile


def get_sample_rate(audio_path: Path) -> int:
    """Gets the current sample rate of the given audio file.

    Parameters:
        audio_path: The path to the audio file.

    Returns:
        The sample rate of the given file.
    """
    with ReadableAudioFile(str(audio_path)) as audio_file:
        return int(audio_file.samplerate)


def resample(audio_path: Path, destination: Path, sample_rate: int) -> None:
    """Copies a wav file to the destination, with the given
    sample rate.

    Parameters:
        audio_path (Path): The path of the file to resample
        destination (Path): The destination at which to create the resampled file
        sample_rate (int): The sample rate for the resampled audio.
    """
    with ReadableAudioFile(str(audio_path)).resampled_to(sample_rate) as audio_file:
        data = audio_file.read(audio_file.frames)
        num_channels = audio_file.num_channels

    with WriteableAudioFile(
        str(destination), samplerate=sample_rate, num_channels=num_channels
    ) as destination_file:
        destination_file.write(data)


def cut(audio_path: Path, destination: Path, start_ms: int, stop_ms: int) -> None:
    """Creates a new wav file at the destination, restricted to the given start
    and stop times.

    Parameters:
        audio_path (Path): The path of the file to resample.
        destination (Path): The destination at which to create the resampled file.
        start_ms (int): The start time in milliseconds to record from.
        stop_ms (int):  The stop time in milliseconds to record to.
    """

    with ReadableAudioFile(str(audio_path)) as audio_file:
        start = int(start_ms * audio_file.samplerate / 1000)
        stop = int(stop_ms * audio_file.samplerate / 1000)

        audio_file.seek(start)
        data = audio_file.read(stop - start)
        num_channels = audio_file.num_channels
        sample_rate = audio_file.samplerate

    with WriteableAudioFile(
        str(destination), samplerate=sample_rate, num_channels=num_channels
    ) as destination_file:
        destination_file.write(data)



from itertools import chain
from pathlib import Path
from typing import List, Optional, Tuple

from loguru import logger
from pympi.Elan import Eaf

from elpis.models import Annotation, ElanOptions, ElanTierSelector


def extract_annotations(
    transcription_file: Path, elan_options: Optional[ElanOptions] = None
) -> List[Annotation]:
    """Extracts annotations from the supplied transcription file.

    If the transcription file is an elan file, elan_options is required.

    Parameters:
        transcription_file: The file from which to extract annotations
        elan_options: Options to include for determining how to extract annotations
                from elan data.

    Returns:
        A list of found annotations.
        Returns an empty list if there was a problem.
    """
    if transcription_file.suffix == ".txt":
        return extract_text_annotations(transcription_file)

    if transcription_file.suffix != ".eaf":
        logger.error(f"Unrecognised file format: {transcription_file}")
        return []

    if elan_options is None:
        logger.error(f"Missing elan options for extraction job.")
        return []

    return extract_elan_annotations(
        transcription_file,
        selection_type=elan_options.selection_mechanism,
        selection_data=elan_options.selection_value,
    )


def extract_text_annotations(file: Path) -> List[Annotation]:
    """Extract transcription information from a text file.

    Parameters:
        file_name: The name of the downloaded file.

    Returns:
        A list of utterance information for the given file.
    """
    with open(file) as transcription_file:
        transcription = transcription_file.read()

    return [
        Annotation(
            audio_file=file.parent / (file.stem + ".wav"),
            transcript=transcription,
        )
    ]


def extract_elan_annotations(
    elan_file_path: Path, selection_type: ElanTierSelector, selection_data: str
) -> List[Annotation]:
    """Extracts annotations from a particular tier in an eaf file (ELAN
    Annotation Format).

    Tiers are nodes from the tree structure in the .eaf file.
    The tier to read from is determined by tier order (eg top tier would be order 1),
    tier type (eg default-lt) or tier name (eg Phrase).

    Parameters:
        elan_file_path: The path to the eaf file.
        selection_type: The method of determining which tier data to extract.
        selection_data: The data corresponding to the selection_type.

    Returns:
        A list of the annotations contained for the supplied data. Returns an
        empty list if the given selection isn't found.
    """
    logger.info(
        f"processing eaf {elan_file_path} using {selection_type}: {selection_data}"
    )

    match selection_type:
        case ElanTierSelector.NAME:
            return get_annotations_by_tier_name(elan_file_path, selection_data)
        case ElanTierSelector.TYPE:
            return get_annotations_by_tier_type(elan_file_path, selection_data)
        case ElanTierSelector.ORDER:
            try:
                order = int(selection_data)
            except:
                order = 1
            return get_annotations_by_tier_order(elan_file_path, order)


def get_annotations_by_tier_order(
    elan_file_path: Path, tier_order: int
) -> List[Annotation]:
    """Retrieves all annotations for a given tier order within an eaf file.

    Parameters:
        elan_file_path: The path to the eaf file.
        tier_order: The tier order to extract from (starts at 1)

    Returns:
        A list of the annotations contained for the supplied tier order.
        Returns an empty list if the given tier order exceeds the nesting of
        the file.
    """
    elan = Eaf(elan_file_path)

    tier_names: List[str] = list(elan.get_tier_names())
    if tier_order > len(tier_names):
        logger.error(
            f"tier_order: {tier_order} exceeds tier length for {elan_file_path}"
        )
        return []

    tier_name = tier_names[tier_order - 1]
    return get_annotations_by_tier_name(
        elan_file_path=elan_file_path, tier_name=tier_name
    )


def get_annotations_by_tier_type(
    elan_file_path: Path, tier_type: str
) -> List[Annotation]:
    """Retrieves all annotations for a given linguistic tier type in an eaf file.

    Parameters:
        elan_file_path: The path to the eaf file.
        tier_type: The linguistic type from which to extract Annotation data.

    Returns:
        A list of the annotations contained for the supplied linguistic type.
        Returns an empty list if the type is not found.
    """
    elan = Eaf(elan_file_path)

    if tier_type not in list(elan.get_linguistic_type_names()):
        logger.error(f"tier_type: {tier_type} not found in file: {elan_file_path}")
        return []

    tier_names = elan.get_tier_ids_for_linguistic_type(tier_type)
    annotations = (
        get_annotations_by_tier_name(elan_file_path, name) for name in tier_names
    )
    # Flatten list of annotations
    return list(chain(*annotations))


def get_annotations_by_tier_name(
    elan_file_path: Path, tier_name: str
) -> List[Annotation]:
    """Retrieves all annotations for a given tier name in an eaf file.

    Parameters:
        elan_file_path: The path to the eaf file.
        tier_name: The tier name from which to extract Annotation data.

    Returns:
        A list of the annotations contained for the supplied tier name.
        Returns an empty list if the name is not found.
    """
    elan = Eaf(elan_file_path)

    if tier_name not in list(elan.get_tier_names()):
        logger.error(f"tier_name: {tier_name} not found in file {elan_file_path}")
        return []

    def create_annotation(elan_annotation: Tuple[str, str, str]) -> Annotation:
        start, end, transcript = elan_annotation
        return Annotation(
            audio_file=elan_file_path.parent / (elan_file_path.stem + ".wav"),
            transcript=transcript,
            start_ms=int(start),
            stop_ms=int(end),
        )

    return list(map(create_annotation, elan.get_annotation_data_for_tier(tier_name)))



import json
import shutil
from copy import copy
from itertools import chain
from pathlib import Path
from typing import Iterable, List, Tuple

from loguru import logger

import elpis.utils.audio as audio
from elpis.datasets.clean_text import clean_text
from elpis.datasets.dataset import CleaningOptions, ProcessingBatch
from elpis.datasets.extract_annotations import extract_annotations
from elpis.models.annotation import Annotation

DEFAULT_DIR = Path("/tmp")
TARGET_SAMPLE_RATE = 16_000


def process_batch(
    batch: ProcessingBatch, output_dir: Path = DEFAULT_DIR
) -> Iterable[Path]:
    """Generates training files from the processing batch and puts them in
    the given directory.

    Parameters:
        batch: The processing batch to generate files from
        output_dir: The directory in which to stick the files.

    Returns:
        The paths of the generated files.
    """
    annotations = extract_annotations(
        transcription_file=batch.transcription_file, elan_options=batch.elan_options
    )

    annotations = map(
        lambda annotation: clean_annotation(annotation, batch.cleaning_options),
        annotations,
    )

    # Generate training files from the annotations
    return chain(
        *map(
            lambda annotation: generate_training_files(
                annotation, output_dir=output_dir
            ),
            annotations,
        )
    )


def clean_annotation(
    annotation: Annotation, cleaning_options: CleaningOptions
) -> Annotation:
    """Cleans the text within an annotation.

    Parameters:
        annotation: The annotation to clean.
        cleaning_options: The cleaning options for the dataset.

    Returns:
        A new annotation whose transcript has been cleaned.
    """
    transcript = clean_text(
        text=annotation.transcript,
        words_to_remove=cleaning_options.words_to_remove,
        punctuation_to_explode=cleaning_options.punctuation_to_explode,
        punctuation_to_remove=cleaning_options.punctuation_to_remove,
    )
    result = copy(annotation)
    result.transcript = transcript
    return result


def generate_training_files(
    annotation: Annotation, output_dir: Path = DEFAULT_DIR
) -> Tuple[Path, Path]:
    """Generates a transcript and audio file pairing for this annotation.

    If the annotation is timed (has a start and stop time), we return a path
    to a new audio file, which is constrained to the given times. Otherwise,
    the annotation spans the entire audio path, and so we return this path,
    unmodified.

    Parameters:
        annotation: The annotation for a given section of audio within the
            supplied audio_file.
        output_dir: The directory in which to store the generated files.

    Returns:
        A tuple containing a transcription and audio file path for the given
            annotation.
    """
    # Get a unique name prefix based on annotation start time
    audio_file = annotation.audio_file
    name = audio_file.stem
    if annotation.start_ms is not None:
        name = f"{name}_{annotation.start_ms}"

    # Save audio file.
    if annotation.is_timed():
        cut_audio_file = output_dir / f"{name}.wav"
        audio.cut(
            audio_path=audio_file,
            destination=cut_audio_file,
            start_ms=annotation.start_ms,  # type: ignore
            stop_ms=annotation.stop_ms,  # type: ignore
        )
        audio_file = cut_audio_file
    else:
        # Make sure we're putting the audio file in the output dir
        if audio_file.parent != output_dir:
            shutil.copy(str(audio_file), str(output_dir / audio_file.name))
            audio_file = output_dir / audio_file.name

    # Resample audio to standardise for training
    audio.resample(
        audio_path=audio_file,
        destination=audio_file,
        sample_rate=TARGET_SAMPLE_RATE,
    )

    # Save gimped transcription_file
    next_annotation = Annotation(
        audio_file=audio_file, transcript=annotation.transcript
    )
    transcription_file = output_dir / f"{name}.json"
    with open(transcription_file, "w") as f:
        json.dump(next_annotation.to_dict(), f)

    return transcription_file, audio_file


def has_finished_processing(
    dataset_files: List[str], processed_files: List[str]
) -> bool:
    """Checks whether the dataset has finished processing.

    Parameters:
        dataset_files: A list of names of the files in the dataset.
        processed_files: A list of names of files uploaded to cloud storage for
            the corresponding dataset.

    Returns:
        true iff the supplied list of processed files would be a valid
        processed dataset for the initial files.
    """

    required_stems = {Path(name).stem for name in dataset_files}
    uploaded_stems = {Path(name).stem for name in processed_files}

    def is_processed(required_stem: str) -> bool:
        starts_with_required_stem = lambda stem: stem.startswith(required_stem)
        return any(map(starts_with_required_stem, uploaded_stems))

    return all(map(is_processed, required_stems))



from pathlib import Path
from typing import Optional

from loguru import logger
from transformers import AutoModelForCTC, AutoProcessor, Trainer

from elpis.datasets import create_dataset, prepare_dataset
from elpis.trainer.data_collator import DataCollatorCTCWithPadding
from elpis.trainer.job import TrainingJob
from elpis.trainer.utils import log_to_file


def train(
    job: TrainingJob,
    output_dir: Path,
    dataset_dir: Path,
    cache_dir: Optional[Path] = None,
    log_file: Optional[Path] = None,
) -> Path:
    """Trains a model for use in transcription.

    Parameters:
        job: Info about the training job, e.g. training options.
        output_dir: Where to save the trained model.
        dataset_dir: A directory containing the preprocessed dataset to train with.
        cache_dir: A directory to use for caching HFT downloads and datasets.
        log_file: An optional file to write training logs to.

    Returns:
        A path to the folder containing the trained model.
    """

    @log_to_file(log_file)
    def _train():
        logger.info("Preparing Datasets...")
        dataset = create_dataset(dataset_dir, cache_dir)
        processor = AutoProcessor.from_pretrained(job.base_model, cache_dir=cache_dir)
        dataset = prepare_dataset(dataset, processor)
        logger.info("Finished Preparing Datasets")

        logger.info("Downloading pretrained model...")
        model = AutoModelForCTC.from_pretrained(
            job.base_model,
            cache_dir=cache_dir,
            ctc_loss_reduction="mean",
            pad_token_id=processor.tokenizer.pad_token_id,
        )
        logger.info("Downloaded model.")

        if job.options.freeze_feature_extractor:
            model.freeze_feature_extractor()

        data_collator = DataCollatorCTCWithPadding(processor=processor, padding=True)
        output_dir.mkdir(exist_ok=True, parents=True)

        trainer = Trainer(
            model=model,
            args=job.to_training_args(output_dir),
            train_dataset=dataset["train"],  # type: ignore
            eval_dataset=dataset["test"],  # type: ignore
            tokenizer=processor.feature_extractor,
            data_collator=data_collator,
        )

        logger.info(f"Begin training model...")
        trainer.train()
        logger.info(f"Finished training!")

        logger.info(f"Saving model @ {output_dir}")
        trainer.save_model()
        trainer.save_state()
        processor.save_pretrained(output_dir)
        logger.info(f"Model written to disk.")

        return output_dir

    return _train()



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional


@dataclass
class Annotation:
    """A class which represents a section of speech for a given audio file and
    sample rate. If start_ms and end_ms aren't specified, it is assumed that the
    Annotation spans the entire audio file.
    """

    audio_file: Path
    transcript: str
    start_ms: Optional[int] = None
    stop_ms: Optional[int] = None

    def is_timed(self) -> bool:
        """Returns true iff the annotation exists between a start and stop time for
        the given recording.
        """
        return self.start_ms is not None and self.stop_ms is not None

    def to_dict(self) -> Dict[str, Any]:
        """Converts an annotation to a serializable dictionary"""
        result = dict(self.__dict__)
        result["audio_file"] = str(self.audio_file)
        return result

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Annotation:
        """Builds an annotation from a serializable dictionary

        Throws an error if the required keys are not found.
        """
        return cls(
            audio_file=Path(data["audio_file"]),
            transcript=data["transcript"],
            start_ms=data.get("start_ms"),
            stop_ms=data.get("stop_ms"),
        )



from typing import List

from pympi.Elan import Eaf

from elpis.models.annotation import Annotation


def build_elan(annotations: List[Annotation], tier_name: str = "Phrase") -> Eaf:
    """Builds an elan file from the given annotations and tier_name.

    Parameters:
        annotations: The list of annotations to add.
        tier_name: The name of the tier to add the annotations under.

    Returns:
        The resulting elan file.
    """
    result = Eaf()
    result.add_tier(tier_name)
    for annotation in annotations:
        result.add_annotation(
            id_tier=tier_name, start=annotation.start_ms, end=annotation.stop_ms
        )
    return result


def build_text(annotations: List[Annotation]) -> str:
    """Combines all the text from a list of annotations, ordered from earliest
    start time.

    Parameters:
        annotations: The list of annotations to Combines

    Returns:
        The combined transcripts.
    """
    annotations = sorted(
        annotations,
        key=(
            lambda annotation: annotation.start_ms
            if annotation.start_ms is not None
            else 0
        ),
    )
    return " ".join(annotation.transcript for annotation in annotations)



from dataclasses import dataclass
from typing import Dict, List, Optional, Union

import torch
from transformers import Wav2Vec2Processor


@dataclass
class DataCollatorCTCWithPadding:
    processor: Wav2Vec2Processor
    padding: Union[bool, str] = True
    max_length: Optional[int] = None
    max_length_labels: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None
    pad_to_multiple_of_labels: Optional[int] = None

    def __call__(
        self, features: List[Dict[str, Union[List[int], torch.Tensor]]]
    ) -> Dict[str, torch.Tensor]:
        # split inputs and labels since they have to be of different lenghts and need
        # different padding methods
        input_features = [
            {"input_values": feature["input_values"]} for feature in features
        ]
        label_features = [{"input_ids": feature["labels"]} for feature in features]

        batch = self.processor.pad(
            input_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )
        with self.processor.as_target_processor():
            labels_batch = self.processor.pad(
                label_features,
                padding=self.padding,
                max_length=self.max_length_labels,
                pad_to_multiple_of=self.pad_to_multiple_of_labels,
                return_tensors="pt",
            )

        # replace padding with -100 to ignore loss correctly
        labels = labels_batch["input_ids"].masked_fill(
            labels_batch.attention_mask.ne(1), -100
        )

        batch["labels"] = labels

        return batch



from __future__ import annotations

from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from elpis.models import ElanOptions

TRANSCRIPTION_EXTENSIONS = {".eaf", ".txt"}


@dataclass
class CleaningOptions:
    """A class representing cleaning options for a dataset."""

    punctuation_to_remove: str = ""
    punctuation_to_explode: str = ""
    words_to_remove: List[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> CleaningOptions:
        kwargs = {field.name: data[field.name] for field in fields(CleaningOptions)}
        return cls(**kwargs)

    def to_dict(self) -> Dict[str, Any]:
        return dict(self.__dict__)


@dataclass
class ProcessingBatch:
    """A class encapsulating the data needed for an individual processing job"""

    audio_file: Path
    transcription_file: Path
    cleaning_options: CleaningOptions
    elan_options: Optional[ElanOptions]

    def to_dict(self) -> Dict[str, Any]:
        result = {}

        result["audio_file"] = str(self.audio_file)
        result["transcription_file"] = str(self.transcription_file)
        result["cleaning_options"] = self.cleaning_options.to_dict()
        if self.elan_options is not None:
            result["elan_options"] = self.elan_options.to_dict()

        return result

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> ProcessingBatch:
        audio_file = Path(data["audio_file"])
        transcription_file = Path(data["transcription_file"])
        cleaning_options = CleaningOptions.from_dict(data["cleaning_options"])
        elan_options = ElanOptions.from_dict(data["elan_options"])
        return cls(
            audio_file=audio_file,
            transcription_file=transcription_file,
            cleaning_options=cleaning_options,
            elan_options=elan_options,
        )


@dataclass
class Dataset:
    """A class representing an unprocessed dataset."""

    name: str
    files: List[Path]
    cleaning_options: CleaningOptions
    elan_options: Optional[ElanOptions]

    def is_empty(self) -> bool:
        """Returns true iff the dataset contains no files."""
        return len(self.files) == 0

    def has_elan(self) -> bool:
        """Returns true iff any of the files in the dataset is an elan file."""
        return any(map((lambda file_name: file_name.suffix == ".eaf"), self.files))

    def is_valid(self) -> bool:
        """Returns true iff this dataset is valid for processing."""
        return (
            not self.is_empty()
            and len(self.files) % 2 == 0
            and len(self.mismatched_files()) == 0
            and len(self.colliding_files()) == 0
        )

    @staticmethod
    def corresponding_audio_name(transcript_file: Path) -> Path:
        """Gets the corresponding audio file name for a given transcript file."""
        return Path(transcript_file).parent / (transcript_file.stem + ".wav")

    def mismatched_files(self) -> Set[Path]:
        """Returns the list of transcript files with no corresponding
        audio and vice versa.

        Corresponding in this case means that for every transcript file with
        name x.some_extension, there is a corresponding file x.wav in the dataset.

        Returns:
            A list of the mismatched file names.
        """
        transcripts_with_audio = set(
            filter(
                lambda file: Dataset.corresponding_audio_name(file) in self.files,
                self._transcript_files(),
            )
        )
        matched_files = transcripts_with_audio | set(
            Dataset.corresponding_audio_name(file) for file in transcripts_with_audio
        )

        return set(self.files).difference(matched_files)

    def colliding_files(self) -> Set[Path]:
        """Returns the list of transcript file names that collide.

        Collide means that two transcript files would be for the same .wav
        file.

        Returns:
            A list of the colliding file names.
        """

        def would_collide(transcript_file: Path) -> bool:
            other_files = self._transcript_files().difference({transcript_file})
            other_file_names = map(lambda file: Path(file).stem, other_files)
            return Path(transcript_file).stem in other_file_names

        return set(filter(would_collide, self._transcript_files()))

    def _transcript_files(self) -> Set[Path]:
        """Returns a set of all transcription files within the dataset."""
        return set(
            filter(lambda file: file.suffix in TRANSCRIPTION_EXTENSIONS, self.files)
        )

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Dataset:
        name = data["name"]
        files = [Path(file) for file in data["files"]]
        cleaning_options = CleaningOptions.from_dict(data["cleaning_options"])

        elan_options = None
        if "elan_options" in data:
            elan_options = ElanOptions.from_dict(data["elan_options"])

        return cls(
            name=name,
            files=files,
            cleaning_options=cleaning_options,
            elan_options=elan_options,
        )

    def to_batches(self) -> List[ProcessingBatch]:
        """Converts a valid dataset to a list of processing jobs, matching
        transcript and audio files.
        """
        return [
            ProcessingBatch(
                transcription_file=transcription_file,
                audio_file=self.corresponding_audio_name(transcription_file),
                cleaning_options=self.cleaning_options,
                elan_options=self.elan_options,
            )
            for transcription_file in self._transcript_files()
        ]

    def to_dict(self) -> Dict[str, Any]:
        result = {
            "name": self.name,
            "files": [file.name for file in self.files],
            "cleaning_options": self.cleaning_options.to_dict(),
        }

        if self.elan_options is not None:
            result["elan_options"] = self.elan_options.to_dict()

        return result



import logging
from functools import wraps
from logging.handlers import WatchedFileHandler
from pathlib import Path
from typing import Callable, Optional

import transformers
from loguru import logger


def log_to_file(log_file: Optional[Path]) -> Callable[[Callable], Callable]:
    """Returns a decorator that will output the logs of a captured function
    to a log file, if it exists.

    Args:
        log_file: The optional file to log to.
    """

    def log_wrapper(func: Callable) -> Callable:
        @wraps(func)
        def function_wrapper(*args, **kwargs):
            if log_file is None:
                return func(*args, **kwargs)

            sink_id = logger.add(log_file)

            # Configure huggingface logging handler
            formatter = logging.Formatter(
                fmt="%(asctime)s.%(msecs)03d | %(levelname)s | (HF) %(name)s |   %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
            handler = WatchedFileHandler(str(log_file))
            handler.setFormatter(formatter)
            handler.setLevel(logging.INFO)

            logging.root.addHandler(handler)
            transformers.logging.set_verbosity_info()
            transformers.logging.enable_propagation()

            result = func(*args, **kwargs)
            # Flush and teardown logging handlers
            logger.remove(sink_id)
            handler.flush()
            logging.root.removeHandler(handler)
            return result

        return function_wrapper

    return log_wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elpis',
 'elpis.datasets',
 'elpis.models',
 'elpis.trainer',
 'elpis.transcriber',
 'elpis.utils']

package_data = \
{'': ['*']}

install_requires = \
['datasets>=2.6.1,<3.0.0',
 'librosa>=0.9.2,<0.10.0',
 'loguru>=0.6.0,<0.7.0',
 'pedalboard>=0.6.2,<0.7.0',
 'pympi-ling>=1.70.2,<2.0.0',
 'torch>=1.12.1,<2.0.0',
 'transformers>=4.23.1,<5.0.0']

setup_kwargs = {
    'name': 'elpis',
    'version': '0.1.4',
    'description': 'A library to perform automatic speech recognition with huggingface transformers.',
    'long_description': '# Elpis Core Library\n\nThe Core Elpis Library, providing a quick api to [:hugs: transformers](https://huggingface.co/models?pipeline_tag=automatic-speech-recognition&sort=downloads)\nfor automatic-speech-recognition.\n\nYou can use the library to:\n\n- Perform standalone inference using a pretrained HFT model.\n- Fine tune a pretrained ASR model on your own dataset.\n- Generate text and Elan files from inference results for further analysis.\n\n## Documentation\n\nDocumentation for the library can be be found [here](https://coedl.github.io/elpis_lib/index.html).\n\n## Dependencies\n\nWhile we try to be as machine-independant as possible, there are some dependencies\nyou should be aware of when using this library:\n\n- Processing datasets (`elpis.datasets.processing`) requires `librosa`, which\n  depends on having `libsndfile` installed on your computer. If you\'re using\n  elpis within a docker container, you may have to manually install\n  `libsndfile`.\n- Transcription (`elpis.transcription.transcribe`) requires `ffmpeg` if your\n  audio you\'re attempting to transcribe needs to be resampled before it can\n  be used. The default sample rate we assume is 16khz.\n- The preprocessing flow (`elpis.datasets.preprocessing`) is free of external\n  dependencies.\n\n## Installation\n\nYou can install the elpis library with:\n`pip3 install elpis`\n\n## Usage\n\nBelow are some typical examples of use cases\n\n### Standalone Inference\n\n```python\nfrom pathlib import Path\n\nfrom elpis.transcriber.results import build_text\nfrom elpis.transcriber.transcribe import build_pipeline, transcribe\n\n# Perform inference\nasr = build_pipeline(pretrained_location="facebook/wav2vec2-base-960h")\naudio = Path("<to_some_audio_file.wav>")\nannotations = transcribe(audio, asr) # Timed, per word annotation data\n\nresult = build_text(annotations) # Combine annotations to extract all text\nprint(result)\n\n# Build output files\ntext_file = output_dir / "test.txt"\nwith open(text_file, "w") as output_file:\n    output_file.write(result)\n```\n\n### Fine-tuning a Pretrained Model on Local Dataset\n\n```python\nfrom pathlib import Path\nfrom typing import List\n\nfrom elpis.datasets import Dataset\nfrom elpis.datasets.dataset import CleaningOptions\nfrom elpis.datasets.preprocessing import process_batch\nfrom elpis.models import ElanOptions, ElanTierSelector\nfrom elpis.trainer.job import TrainingJob, TrainingOptions\nfrom elpis.trainer.trainer import train\nfrom elpis.transcriber.results import build_elan, build_text\nfrom elpis.transcriber.transcribe import build_pipeline, transcribe\n\nfiles: List[Path] = [...] # A list of paths to the files to include.\n\ndataset = Dataset(\n    name="dataset",\n    files=files,\n    cleaning_options=CleaningOptions(), # Default cleaning options\n    # Elan data extraction info- required if dataset includes .eaf files.\n    elan_options=ElanOptions(\n        selection_mechanism=ElanTierSelector.NAME, selection_value="Phrase"\n    ),\n)\n\n# Setup\ntmp_path = Path(\'...\')\n\ndataset_dir = tmp_path / "dataset"\nmodel_dir = tmp_path / "model"\noutput_dir = tmp_path / "output"\n\n# Make all directories\nfor directory in dataset_dir, model_dir, output_dir:\n    directory.mkdir(exist_ok=True, parents=True)\n\n# Preprocessing\nbatches = dataset.to_batches()\nfor batch in batches:\n    process_batch(batch, dataset_dir)\n\n# Train the model\njob = TrainingJob(\n    model_name="some_model",\n    dataset_name="some_dataset",\n    options=TrainingOptions(epochs=2, learning_rate=0.001),\n    base_model="facebook/wav2vec2-base-960h"\n)\ntrain(\n    job=job,\n    output_dir=model_dir,\n    dataset_dir=dataset_dir,\n)\n\n# Perform inference with pipeline\nasr = build_pipeline(\n    pretrained_location=str(model_dir.absolute()),\n)\naudio = Path("<to_some_audio_file.wav>")\nannotations = transcribe(audio, asr)\n\n# Build output files\ntext_file = output_dir / "test.txt"\nwith open(text_file, "w") as output_file:\n    output_file.write(build_text(annotations))\n\nelan_file = output_dir / "test.eaf"\neaf = build_elan(annotations)\neaf.to_file(str(elan_file))\n\nprint(\'voila ;)\')\n```\n',
    'author': 'Harry Keightley',
    'author_email': 'harrykeightley@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/CoEDL/elpis_lib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from elpis.datasets.dataset import CleaningOptions, Dataset, ProcessingBatch
from elpis.datasets.preprocessing import process_batch
from elpis.datasets.processing import create_dataset, prepare_dataset

__all__ = [
    "CleaningOptions",
    "Dataset",
    "ProcessingBatch",
    "process_batch",
    "create_dataset",
    "prepare_dataset",
]
